void EmptyBGMap (unsigned int WhichOne)
{
  // Sets all cells in a BGMap to 0.
  
  SetMem((void*)(0x20000 + WhichOne * 0x2000), 0, 8192);
}

void FadeIn ()
{
  register unsigned int i;

  for(i = 0; i <= Brightness; i += 2)
  {
    WaitForVIP();
    DoSounds();
    VIP_REGS[BRTA] = i;
    VIP_REGS[BRTB] = i << 1;
    VIP_REGS[BRTC] = i;
  }
}

void FadeOut ()
{
  register signed int i;

  for(i = Brightness; i >= 0; i -= 2)
  {
    WaitForVIP();
    DoSounds();
    VIP_REGS[BRTA] = i;
    VIP_REGS[BRTB] = i << 1;
    VIP_REGS[BRTC] = i;
  }
}

void LoadEfficientBGMap (
  const u16* Src, unsigned int BGMap, unsigned int CharIndexBase
)
{
  // Loads an efficiently stored vbimgc-produced BGMap and optionally adjusts
  // the character indexes of its cells.

  register unsigned int X;
  register unsigned int Y;
  register unsigned int Width;
  register unsigned int Height;
  register unsigned int Index;

  Width = Src[0] & 0x00FF;
  Height = Src[0] >> 8;

  Index = 1;
  for(Y = 0; Y < Height; Y++)
    for(X = 0; X < Width; X++)
      BGMM[BGMap * 0x1000 + Y * 64 + X] = Src[Index++] + CharIndexBase;
}

unsigned int LoadEfficientCharset (const u16* Src, unsigned int CharIndex)
{
  // Loads an efficiently stored vbimgc-produced charset and returns the
  // index of the character immediately following it in the VIP's character
  // memory.

  register unsigned int i;
  register unsigned int Length;

  if(CharIndex >= 2048)
    Crash("LOADEFFICIENTCHARSET: CHARINDEX!");

  CharIndex *= 8;  // Convert to row index

  Length = Src[0] * 8;
  for(i = 1; i <= Length; i++)
    vbCharmem[CharIndex++] = Src[i];

  return CharIndex / 8;
}

void SetBGMapPalette (unsigned int Index, unsigned int Palette)
{
  // Modifies all cells of a BGMap to use the specified palette, which must
  // be a BGM_PAL* value.

  register unsigned int i;
  register u16* BGMapPtr;

  BGMapPtr = (u16*)(0x20000 + Index * 0x2000);

  for(i = 0; i < 64 * 64; i++)
    BGMapPtr[i] = (BGMapPtr[i] & 0x3FFF) | Palette;
}

void SetEndWorld (unsigned int World)
{
  // Sets the specified world to be the end world.

  vbWorld[World].Flags = WRLD_END;
}

void WaitForVIP ()
{
  // Waits for the VIP to finish drawing the current frame.

  while(!(VIP_REGS[INTPND] & XPEND));
  VIP_REGS[INTCLR] = XPEND;
}
