// Auto pause

#define kAutoPausePeriod (15 * 60 * 50)  // How often (15 minutes)
tyBoolean AutoPauseOn;  // Whether it's enabled
u32 AutoPauseFrames;  // Frames until next auto pause

// Floor

#define kFloorH 112  // Floor height

s16 FloorBeginningMY;  // the MY where the current level's floor begins
s16 FloorEndingMY;  // the MY where it ends (= FloorBeginningMY + (8 << 3))
s16 FloorMYOffset;  // the offset - decreases to provide a feeling of speed
s16 FloorMYOffsetDelta;  // how much it decreases per video frame
s16 FloorMXOffset;  // used to pan the floor left and right

// Things

#define kMaxThings 28

tyThing Thing[kMaxThings];
tyThing* ViewThing[kMaxThings + 1];  // Things sorted by Z value for rendering

#define kThkNothing     0  // Nothing (slot empty)
#define kThkTree        1  // Tree
#define kThkWaitingTree 2  // Tree waiting to be respawned (not on-screen)
#define kThkBiker       3  // Enemy biker
#define kThkTank        4  // Enemy tank
#define kThkMissile     5  // Player's missile (only one at a time)

#define kMinWidth int2fix(5)
#define kMaxWidth int2fix(120)

// Min/max number of frames a tree should wait before respawning.
#define kTreeRespawnMin 2
#define kTreeRespawnMax 25

// Min/max number of frames before spawning a tank.
#define kTankRespawnMin 1000
#define kTankRespawnMax 2000

// Biker X coordinate limits (screen coordinates; use with RevProjX)
#define kMinBikerX 32
#define kMaxBikerX (383 - 32)

// BGMap numbers for different classes of things
#define kMNMain 0
#define kMNStatus 1
#define kMNScratch 2
#define kMNDayTrees 3
#define kMNFloor 14

// Speeds
const fix14 Throttle[5] = {
  float2fix(0.0),
  float2fix(0.2),
  float2fix(0.4),
  float2fix(0.6),
  float2fix(0.8),
};

// Game state
u8 PlayerSpeed; // 0 to 4; 0 = stopped
signed char PlayerHeading; // -1 = left, 0 = straight ahead, 1 = right
unsigned int Score;
unsigned int HighScore;
u8 Level;  // Level number (1 to 8) * 2; lowest bit indicates night
u8 NumBikers;  // How many enemy bikers are still alive
tyBoolean TankExists;  // Whether a tank is present (can only be one at a time)
unsigned int TankFrames;  // How many frames until we spawn a tank
tyThing* Missile;  // Pointer to missile (NULL if not present)

// This gets multiplied by the Throttle values to obtain, essentially, a pixel
// displacement
#define TurnSpd float2fix(80)

#define IsTurningN (PlayerHeading == 0)
#define IsTurningL (PlayerHeading & 0x80)
#define IsTurningR (PlayerHeading & 0x7F)

#define TurnN (PlayerHeading = 0)
#define TurnL (PlayerHeading = -1)
#define TurnR (PlayerHeading = 1)

// Used in ProcessThings
#define PTTurnMacro (IsTurningL ? int2fix(72) : (IsTurningR ? -int2fix(72) : 0))

// 3D projection

#define kHorizonZ float2fix(64.0)

// Parallax limit; always good to avoid blinding your audience ;) -AB
fix14 kPlxLimit = int2fix(15);

// Calculate screen X, Y, and P coords from world X, Y, and Z coords
#define ProjectX(x,z) (192 + fix2int(fixDiv((x), (z))))
#define ProjectY(y,z) (112 + fix2int(fixDiv((y), (z))))
#define ProjectP(z) fixMul(fixDiv(z, kHorizonZ), kPlxLimit)

// Solving for X:
// X' = 192 + (X / Z)
// X' - 192 = X / Z
// X = (X' - 192) * Z

// Reverse project a given screen point (integer) into a world point
// (fixed-point) on a given Z plane (fixed-point)
#define RevProjX(xp,z) fixMul(int2fix((xp) - 192), (z))
#define RevProjY(yp,z) fixMul(int2fix((yp) - 112), (z))

// Limits for spawning
#define kLeftXLimit RevProjX(96, int2fix(10))
#define kRightXLimit RevProjX(383 - 96, int2fix(10))

void WaitForStartOrA ()
{
  // Waits for the player to press and release Start or A.
  // While waiting, the random seed is incremented.

  while(!(vbReadPad() & (K_STA | K_A)))
    RandomSeed++;
  while(vbReadPad() & (K_STA | K_A))
    RandomSeed++;
}

void PrecautionScreen ()
{
  EmptyBGMap(0);

  vbTextOut(0, 13, 9,  "IMPORTANT:");
  vbTextOut(0, 13, 11, "READ INSTRUCTION AND");
  vbTextOut(0, 13, 13, "PRECAUTION BOOKLETS");
  vbTextOut(0, 13, 15, "BEFORE OPERATING");

  vbSetWorld(31, WRLD_ON | 0, 0, 0, 0, 0, 0, 0, 383, 223);
  SetEndWorld(30);

  vbDisplayShow();
  WaitForStartOrA();
  FadeOut();
}

void AdjustmentScreen ()
{
  LoadEfficientBGMap(kBGM_AdjustmentLR, 0, 1);

  LoadEfficientCharset(kChr_Adjustment, 1);

  vbSetWorld(31, WRLD_LON | 0, 0, 0, 0, 0, 0, 0,   383, 223);
  vbSetWorld(30, WRLD_RON | 0, 0, 0, 0, 0, 0, 224, 383, 223);
  SetEndWorld(29);

  FadeIn();
  WaitForStartOrA();
  FadeOut();
}

void AutoPauseSelect ()
{
  // Kludged to eleven! -MH

  register tyBoolean State;
  register u16 Buttons;
  register u16 ButtonsOld;
  register u16 ButtonsDiff;

  EmptyBGMap(0);

  vbTextOut(0, 17, 11, "AUTOMATIC PAUSE");
  vbTextOut(0, 17, 13, "   ON     OFF  ");

  vbSetWorld(31, WRLD_ON | 0, 0, 0, 0, 0, 0, 0, 383, 223);
  SetEndWorld(30);

  FadeIn();

  ButtonsOld = vbReadPad();
  State = true;
  PlaySound(sndMenuBeep);
  for(;;)
  {
    Buttons = vbReadPad();
    ButtonsDiff = (Buttons ^ ButtonsOld) & Buttons;
    ButtonsOld = Buttons;
    if (ButtonsDiff & ((State ? K_LR : K_LL) | K_SEL))
    {
      PlaySound(sndMenuBeep);
      State = !State;
    }

    if (ButtonsDiff & (K_STA | K_A))
      break;

    WaitForVIP();
    vbCharOut(0, 19, 13, State ? '<' : ' ');
    vbCharOut(0, 22, 13, State ? '>' : ' ');
    vbCharOut(0, 26, 13, State ? ' ' : '<');
    vbCharOut(0, 30, 13, State ? ' ' : '>');

    if(Buttons & (K_STA | K_A))
      break;

    DoSounds();

    RandomSeed++;
  }

  PlaySound(sndMenuConfirm);

  FadeOut();

  AutoPauseOn = State;
}

void AutoPause ()
{
  // Shows the automatic pause screen if appropriate.

  // Do nothing if auto pause is off or if the counter hasn't reached zero.
  if(!AutoPauseOn || AutoPauseFrames)
    return;

  EmptyBGMap(kMNScratch);

  vbTextOut(kMNScratch, 17, 13, "AUTOMATIC PAUSE");

  vbSetWorld(31, WRLD_ON | kMNScratch, 0, 0, 0, 0, 0, 0, 383, 223);
  SetEndWorld(30);

  FadeIn();
  WaitForStartOrA();
  FadeOut();

  AutoPauseFrames = kAutoPausePeriod;
}

void TitleScreen ()
{
  LoadEfficientBGMap(kBGM_TitleLR, 0, 0);

  LoadEfficientCharset(kChr_Title, 0);

  vbSetWorld(31, WRLD_LON | 0, 0, 0, 0, 0, 0, 0, 383, 223);
  vbSetWorld(30, WRLD_RON | 0, 0, 0, 0, 0, 0, 224, 383, 223);
  SetEndWorld(29);

  FadeIn();
  WaitForStartOrA();
  PlaySound(sndMenuConfirm);
  FadeOut();
}

void GameOver ()
{
  // Note: this function assumes that GameLoop has already loaded its
  // charsets and BGMaps.

  register vbtyAffinePar* Line;
  register unsigned int i;
  register s16 MX;  // 13.3 fixed point
  register s16 MY;  // 13.3 fixed point

  EmptyBGMap(kMNScratch);

  vbSetWorld(31, WRLD_ON | WRLD_AFFINE | kMNFloor, 0, 0, 0, 0, 0, 0, 383, 223);
  vbWorld[31].ParAddr = 0x1000;  // use BGMap 1 as parameter space
  vbSetWorld(30, WRLD_ON | kMNScratch, 160, -8, 64, 0, 0, 0, 200, 79);  // Nums
  vbSetWorld(29, WRLD_ON | kMNMain, 64, -8, 64, 160, 0, 192, 55, 79);  // Text
  SetEndWorld(28);

  if(Score > HighScore)
    HighScore = Score;

  PrintU32Big(kMNScratch, 0, 0, Level >> 1);
  PrintU32Big(kMNScratch, 0, 4, Score);
  PrintU32Big(kMNScratch, 0, 8, HighScore);

  MX = 0;
  vbDisplayShow();
  for(;;)
  {
    RandomSeed++;

    WaitForVIP();

    Line = (vbtyAffinePar*)(0x22000);  // use BGMap 1 as parameter space

    MY = (4 * 8) << 3;
    for(i = 0; i < 224; i++)  // I don't care if this runs at 25 Hz. -MH
    {
      Line->MX = MX + (i & 7);
      Line->MP = 0;
      Line->MY = MY;
      Line->XIncr = (1 << 9) / 8;
      Line->YIncr = 0;

      MY++;
      if(MY >= ((5 * 8) << 3))
        MY -= 8 << 3;

      Line++;  // Advance to next affine line
    }

    if(vbReadPad() & (K_A | K_STA))
      break;

    MX++;
  }
  FadeOut();
}

void SetFloor (unsigned int Which)
{
  // Sets the floor texture.

  while(Which > 4) Which -= 4;

  FloorBeginningMY = (Which * 8) << 3;
  FloorEndingMY = FloorBeginningMY + (8 << 3);
}

void RenderFloor ()
{
  // Calculates the affine parameters for the floor.

  // This is all a big hack so don't expect to find much wisdom here.

  // BGMap 14 begins at 0x3C000. We use 5 rows (0x280 bytes) of it.
  // The parameter array begins at 0x3D100 (assuming kFloorH = 112).

  register vbtyAffinePar* Line;
  register unsigned int i;
  register s16 MY;  // 13.3 fixed point
  register s16 MYIncr;  // 13.3 fixed point
  register s16 XIncr;  // 7.9 fixed point

  static const s16 kXIncr[kFloorH] = {
    #include "floorxi.h"
  };

  Line = (vbtyAffinePar*)(0x3D800 - kFloorH * sizeof(vbtyAffinePar));

  MYIncr = 16 /* obtained by experiment */;
  MY = FloorBeginningMY + FloorMYOffset;
  for(i = 0; i < kFloorH; i++)
  {
    XIncr = kXIncr[i];

    Line->MX = -(((384 / 2) * XIncr) >> 6) - FloorMXOffset;
    Line->MP = kXIncr[kFloorH - i] >> 6 /* obtained by experiment */;
    Line->MY = MY;
    Line->XIncr = XIncr;
    Line->YIncr = 0;

    if(!(i & 31 /* obtained by experiment */))
      MYIncr /= 2;
    MY += MYIncr;
    if(MY >= FloorEndingMY)
      MY = FloorBeginningMY;

    Line++;  // Advance to next affine line
  }
}

tyThing* CreateThing (fix14 X, fix14 Z, u8 Kind)
{
  // Creates a thing with the specified properties and returns a pointer to it.

  register unsigned int i;
  register tyThing* Th;

  for(i = 0; i < kMaxThings; i++)
    if((Th = &(Thing[i]))->Kind == kThkNothing)
    {
      Th->X = X;
      Th->Z = Z;
      Th->Kind = Kind;
      return Th;
    }

  // Should never happen - see RandomizeThings
  Crash("CREATETHING: NO MORE THINGS");
  return NULL;  // Crash never returns anyway
}

void PlayerControl ()
{
  register u16 Buttons;
  register unsigned int OldPlayerSpeed;

  // Forces the player to release A/B each time when changing speed.
  static tyBoolean ABDown;

  Buttons = vbReadPad();

  if(Buttons & K_LL)
  {
    if(PlayerSpeed)
    {
      FloorMXOffset += PlayerSpeed;
      TurnL;
    }
    vbWorld[1].MX = 72;  // Show the motorcycle going left
    if(Missile) Missile->X -= int2fix(50);
  }
  else if(Buttons & K_LR)
  {
    if(PlayerSpeed)
    {
      FloorMXOffset -= PlayerSpeed;
      TurnR;
    }
    vbWorld[1].MX = 72 + 72;  // Show the motorcycle going right
    if(Missile) Missile->X += int2fix(50);
  }
  else if((Buttons & (K_LR | K_LL)) == 0)
  {
    vbWorld[1].MX = 0;  // Show the motorcycle going straight
    TurnN;
  }

  if((Buttons & (K_A | K_B)) && ABDown)
    return;
  ABDown = Buttons & (K_A | K_B);
  OldPlayerSpeed = PlayerSpeed;
  if(Buttons & K_A)
    if(++PlayerSpeed > 4) PlayerSpeed = 4;
  if(Buttons & K_B)
    if(PlayerSpeed) PlayerSpeed--;

  if(PlayerSpeed != OldPlayerSpeed)
  {
    FloorMYOffsetDelta = PlayerSpeed * 2;  // KLUDGE!!!
    OldPlayerSpeed = PlayerSpeed * 0x80;  // Variable reuse :P - frequency
    SND_REGS[WAVE4].SxFQH = OldPlayerSpeed >> 8;
    SND_REGS[WAVE4].SxFQL = OldPlayerSpeed & 0xFF;
  }

  if((Buttons & (K_LT | K_RT)) && !Missile)
  {
    // Blam, blam, blam!
    Missile = CreateThing(0, int2fix(1), kThkMissile);
    PlaySound(sndMissileLaunch);
  }

  #if kDebug
    if(Buttons & K_SEL)
      // Force advance to next level
      NumBikers = 0;
    if(Buttons & K_STA)
      // Force automatic pause when entering next level
      AutoPauseFrames = 0;
  #endif
}

inline unsigned int CalcTreeMX (fix14 Z)
{
  // Calculates the MX value for a tree's world.

  if(Z <= float2fix(4.0)) return 224;
  if(Z <= float2fix(4.5)) return 130;
  if(Z <= float2fix(5.0)) return 56;
  if(Z <= float2fix(6.0)) return 0;
  if(Z <= float2fix(8.5)) return 360;
  if(Z <= float2fix(12.0)) return 227;
  if(Z <= float2fix(16.0)) return 139;
  if(Z <= float2fix(24.0)) return 74;
  if(Z <= float2fix(32.0)) return 40;
  if(Z <= float2fix(48.0)) return 22;
  return 9;
}

inline unsigned int CalcTreeWidth (fix14 Z)
{
  // Calculates the width value for a tree's world.

  if(Z <= float2fix(4.0)) return 119;
  if(Z <= float2fix(4.5)) return  89;
  if(Z <= float2fix(5.0)) return  71;
  if(Z <= float2fix(6.0)) return  55;
  if(Z <= float2fix(8.5)) return 119;
  if(Z <= float2fix(12.0)) return 89;
  if(Z <= float2fix(16.0)) return 65;
  if(Z <= float2fix(24.0)) return 41;
  if(Z <= float2fix(32.0)) return 24;
  if(Z <= float2fix(48.0)) return 11;
  return 7;
}

inline unsigned int CalcTreeHeight (fix14 Z)
{
  // Calculates the height value for a tree's world.

  if(Z <= float2fix(6.0)) return 223;
  if(Z <= float2fix(8.5)) return 168;
  if(Z <= float2fix(12.0)) return 99;
  if(Z <= float2fix(16.0)) return 79;
  if(Z <= float2fix(24.0)) return 60;
  if(Z <= float2fix(32.0)) return 33;
  if(Z <= float2fix(48.0)) return 15;
  return 7;
}

void SortThings ()
{
  // Sorts things for rendering. The algorithm is an insertion sort.

  // If this isn't fast enough, increment a counter of non-kThkNothing things
  // in step 1 and use it instead of kMaxThings for the limit of the loop in
  // step 2. Also terminate the ViewThing array with NULL at the end of step 1
  // so that RenderThings will know when to stop, since the ViewThing array may
  // then be shorter than the Thing array.
  // If it STILL isn't fast enough, find out the GAS structure accessing syntax
  // and rewrite the whole thing in assembly. :P
  // -MH

  register unsigned int i;
  register unsigned int j;
  register tyThing** VT;  // Using this produces 30 bytes less machine code :)
  register tyThing* Th;

  VT = ViewThing;

  // Initialize the array (step 1).
  Th = &(Thing[0]);
  for(i = 0; i < kMaxThings; i++)
    VT[i] = Th++;

  // Sort it (step 2).
  for(i = 1; i < kMaxThings; i++)
  {
    Th = VT[i];
    j = i;
    while(j && VT[j - 1]->Z < Th->Z)
    {
      VT[j] = VT[j - 1];
      j--;
    }
    VT[j] = Th;
  }
}

void RenderThings ()
{
  // Calculates things' positions and sizes onscreen, sorts them by Z, and
  // sets their worlds.

  unsigned int WorldNum;
  unsigned int ThingNum;
  unsigned int MapNum;
  unsigned int GX;
  unsigned int GP;
  unsigned int GY;
  unsigned int MX;
  unsigned int MP;
  unsigned int MY;
  unsigned int Width;   // World width minus 1 (VIP requires that)
  unsigned int Height;  // World height minus 1 (VIP requires that)
  register tyThing* Th;

  SortThings();

  WorldNum = 29;
  for(ThingNum = 0; ThingNum < kMaxThings; ThingNum++)
  {
    Th = ViewThing[ThingNum];

    // Avoid division by zero and other shenanigans
    if(Th->Z <= 0)
      continue;

    switch(Th->Kind)
    {
      case kThkNothing:
      case kThkWaitingTree:
        continue;
        break;

      case kThkTree:
        MapNum = kMNDayTrees;
        Width = CalcTreeWidth(Th->Z);
        Height = CalcTreeHeight(Th->Z);
        GX = ProjectX(Th->X, Th->Z) - Width / 2;
        GP = ProjectP(Th->Z);
        GY =
          (Th->Z <= float2fix(8.5)) ?
          0 : ProjectY(int2fix(kFloorH), Th->Z) - Height;
        MX = CalcTreeMX(Th->Z);
        MP = 0;
        MY = (Th->Z <= float2fix(6.0)) ? 288 : 0;
        break;

      case kThkBiker:
        MapNum = kMNMain;
        Width = (PlayerSpeed >= 3) ? 15 : 7;
        Height = (PlayerSpeed >= 3) ? 31 : 15;
        GX = ProjectX(Th->X, Th->Z) - Width / 2;
        GP = ProjectP(Th->Z);
        GY = kFloorH - 7;
        if(PlayerSpeed >= 3)
        {
          MX = (Th->XIncr > 0) ? 160 : 176;
          MY = 48;
        }
        else
        {
          MX = (Th->XIncr > 0) ? 176 : 184;
          MY = 32;
        }
        MP = 0;
        break;

      case kThkTank:
        MapNum = kMNMain;
        Width = 15;
        Height = 7;
        GX = ProjectX(Th->X, Th->Z) - Width / 2;
        GP = ProjectP(Th->Z);
        GY = kFloorH - 8;
        MX = 160;
        MY = 40;
        MP = 0;
        break;

      case kThkMissile:
        MapNum = kMNMain;
        Width = 7;
        Height = 15;
        GX = ProjectX(Th->X, Th->Z) - Width / 2;
        GP = ProjectP(Th->Z);
        GY = ProjectY(int2fix(kFloorH), Th->Z);  //ProjectY(0, Th->Z);
        MX = 208;
        MY = 80;
        MP = 0;
        break;
   }

    // Set up the world
    vbSetWorld(
      WorldNum--,
      WRLD_ON | MapNum,
      GX,
      GP,
      GY,
      MX,
      MP,
      MY,
      Width,
      Height
    );
  }

  // Hide unused thing worlds
  while(WorldNum > 1)
    vbWorld[WorldNum--].Flags = 0;
}

tyBoolean MissileLogic ()
{
  // Moves the missile forward and checks whether it collides with something.
  // Returns true if the missile is to be destroyed.

  register unsigned int i;
  register fix14 Diff;
  register tyThing* Th;

  for(i = 0; i < kMaxThings; i++)
  {
    Th = &(Thing[i]);
    if(Missile->Z < Th->Z)
      continue;

    Diff = Abs(Th->X - Missile->X);

    switch(Th->Kind)
    {
      case kThkTree:
        if(Diff <= int2fix(50))
          return true;
        break;

      case kThkBiker:
        if(PlayerSpeed == 4 && Diff <= int2fix(100))
        {
          Th->Kind = kThkNothing;
          PlayNoiseGradient(50, 1000, 100, 0x88, 0x44);
          Score += 1000;
          NumBikers--;
          return true;
        }
        break;

      case kThkTank:
        if(Diff <= int2fix(200))
        {
          Th->Kind = kThkNothing;
          TankExists = false;
          PlayNoiseGradient(100, 1500, 100, 0xAA, 0x44);
          Score += 5000;
          return true;
        }
        break;
    }
  }

  Missile->Z += float2fix(0.5);
  return Missile->Z > kHorizonZ;
}

void ProcessThings ()
{
  // Handles thing movement, spawning, etc.

  register tyThing* Th;
  register unsigned int i;

  for(i = 0; i < kMaxThings; i++)
  {
    Th = &(Thing[i]);

    switch(Th->Kind)
    {
      case kThkTree:
        // Bring it closer to the player.
        Th->Z -= Throttle[PlayerSpeed];

        // Move it left/right (simulate steering)
        if(IsTurningL)
          Th->X += float2fix(10) * PlayerSpeed;
        else if(IsTurningR)
          Th->X -= float2fix(10) * PlayerSpeed;

        // If it's too close, remove it and set a respawn timer
        if(
          (Th->Z <= int2fix(-1)) ||
          (ProjectX(Th->X, Th->Z) < -7) ||
          (ProjectX(Th->X, Th->Z) > 390)
        )
        {
          Th->Kind = kThkWaitingTree;
          Th->Timer = Random(kTreeRespawnMin, kTreeRespawnMax);
          // Whoosh!
          PlayNoiseGradient(
            25 - PlayerSpeed * 2,
            Random(1500, 2000),
            Random(200, 250),
            (Th->X > 0) ? 0x1C : 0xC1,
            0
          );
        }

        break;

      case kThkWaitingTree:
        // Process its respawn timer - but only if we're moving!
        if(Th->Timer && PlayerSpeed)
          Th->Timer--;
        else
        {
          // Respawn the tree
          Th->X = fixRnd(kLeftXLimit - PTTurnMacro, kRightXLimit - PTTurnMacro);
          Th->Z = kHorizonZ;
          Th->Kind = kThkTree;
        }

        break;

      case kThkBiker:
        Th->Z = kHorizonZ - int2fix(4 * PlayerSpeed);
        if(!(--(Th->Timer)))
        {
          // Change direction and speed periodically.
          Th->XIncr = fixRnd(int2fix(25), int2fix(50));
          if(BooleanRandom())
            Th->XIncr = -(Th->XIncr);
          Th->Timer = Random(50, 150);
        }
        Th->X += Th->XIncr;
        if(
          Th->X < RevProjX(kMinBikerX, Th->Z) ||
          Th->X > RevProjX(kMaxBikerX, Th->Z)
        )
          Th->XIncr = -(Th->XIncr);

        break;

      case kThkTank:
        Th->X += int2fix(40);
        if(Th->X > RevProjX(383, kHorizonZ))
          Th->Kind = kThkNothing;

        break;

      case kThkMissile:
        break;
    }
  }

  // Handle tank respawning
  if(!(--TankFrames))
  {
    if(!TankExists)
    {
      CreateThing(RevProjX(-16, kHorizonZ), kHorizonZ, kThkTank);
      TankExists = true;
    }
    TankFrames = Random(kTankRespawnMin, kTankRespawnMax);
  }

  // Handle the missile if it exists
  if(Missile)
    if(MissileLogic())
    {
      Missile->Kind = kThkNothing;
      Missile = NULL;
    }
}

tyBoolean PlayerCrashed ()
{
  // Checks whether the player has crashed into a tree.

  register unsigned int i;
  register signed int OffCenter;
  register unsigned int AbsOffCenter;
  register tyThing* Th;

  for(i = 0; i < kMaxThings; i++)
  {
    Th = &(Thing[i]);
    if(Th->Kind == kThkTree)
      if(Th->Z < float2fix(3))
      {
        OffCenter = ProjectX(Th->X, Th->Z) - 192;
        AbsOffCenter = Abs(OffCenter);
        if(AbsOffCenter <= 100)
          return true;  // BOOM!
      }
  }

  return false;
}

void RandomizeThings ()
{
  // At level initialization (level change or death), randomly spawns trees,
  // creates enemy bikers, etc.

  unsigned int i;
  unsigned int j;
  unsigned int TreeLimit;
  register tyThing* Th;

  // The following formula will spawn 3 to 24 trees. We have to leave 4 things
  // free for the projectile, tank, and two bikers.
  TreeLimit = (Level >> 1) * 3;
  if(TreeLimit > (kMaxThings - 4))
    TreeLimit = kMaxThings - 4;

  // Spawn trees
  for(i = 0; i < TreeLimit; i++)
  {
    Th = &(Thing[i]);

#if 1
    Th->X = fixRnd(kLeftXLimit, kRightXLimit);
    // To prevent a wall of trees:
    for(;;)
    {
      Th->Z = fixRnd(fixMul(kHorizonZ, float2fix(0.8)), kHorizonZ);
      if(Th->Z > int2fix(16))
        break;
    }
    Th->Kind = kThkTree;
#else
    // Just schedule them to be respawned and ProcessThings will take care of
    // them.
    // Problem with this method: trees seem to come in batches.
    Th->Kind = kThkWaitingTree;
    Th->Timer = Random(kTreeRespawnMin, kTreeRespawnMax);
#endif
  }

  // Spawn two enemies on motorcycles
  for(j = 0; j < 2; j++)
  {
    Th = &(Thing[i++]);
    Th->X = 0;
    Th->Kind = kThkBiker;
    Th->Timer = 1;  // That will take care of the rest (see ProcessThings)
  }
  NumBikers = 2;

  // Clear the rest if any left
  while(i < kMaxThings)
    Thing[i++].Kind = kThkNothing;

  // Schedule the first spawning of a tank
  TankFrames = Random(kTankRespawnMin, kTankRespawnMax);

  // No missile present yet
  Missile = NULL;

  // No tank yet either
  TankExists = false;
}

void InitLevel ()
{
  // Initializes a level. Called when entering a new level or lost a life.

  register SOUNDREG* Wave4;

  // Stop engine noise
  SND_REGS[WAVE4].SxINT = 0;

  PlayerSpeed = 0;
  FloorMYOffset = 0;
  FloorMYOffsetDelta = 0;
  FloorMXOffset = 0;

  SetFloor(Level >> 1);

  RandomizeThings();

  WaitForVIP();

  RenderThings();
  RenderFloor();
  vbWorld[31].Flags = (Level & 1) ? WRLD_ON : kMNMain;  // Moon (BGMap 0)

  Brightness = (Level & 1) ? 16 : 32;

  FadeIn();

  // Prepare channel for engine noise
  Wave4 = &(SND_REGS[WAVE4]);
  Wave4->SxINT = 0x80;
  Wave4->SxFQH = 0;
  Wave4->SxFQL = 0;
  Wave4->SxLRV = 0x44;
  Wave4->SxEV0 = 0xF0;
  Wave4->SxEV1 = 0x02;
}

void InitGameplayWorlds ()
{
  // Sets up worlds for use by GameLoop and functions that it calls.

  register unsigned int i;

  // Moon (off by default; set by InitLevel)
  vbSetWorld(31, kMNStatus, 288, 12, 32, 160, 0, 80, 31, 31);
  // Floor
  vbSetWorld(
    30, WRLD_ON | WRLD_AFFINE | kMNFloor, 0, 0, kFloorH, 0, 0, 0, 383, 111
  );
  vbWorld[30].ParAddr = (u16)(0xEC00 - kFloorH * (sizeof(vbtyAffinePar) / 2));
  // Worlds reserved for things
  for(i = 2; i <= 29; i++) vbWorld[i].Flags = 0;
  // Player's bike
  vbSetWorld(1, WRLD_ON | kMNMain, 156, -8, 192, 0, 0, 0, 71, 31);
  // Status bar
  vbSetWorld(0, WRLD_ON | kMNStatus, 0, -10, 224 - 16, 0, 0, 0, 383, 15);
}

void LevelStartScreen ()
{
  // Note: this function assumes that GameLoop has already loaded its
  // charsets and BGMaps.

  register unsigned int i;

  EmptyBGMap(kMNScratch);

  // Level number
  vbSetWorld(31, WRLD_ON | kMNScratch, 64, -8, 64, 0, 0, 0, 200, 15);
  // "Level" text
  vbSetWorld(30, WRLD_ON | kMNMain, 128, -8, 64, 160, 0, 192, 55, 15);
  // "At night" text
  vbSetWorld(29, (Level & 1) ? WRLD_ON : 0, 160, -8, 128, 160, 0, 176, 55, 15);
  SetEndWorld(28);

  PrintU32Big(kMNScratch, 0, 0, Level >> 1);

  FadeIn();
  for(i = 0; i < 50; i++)
    WaitForVIP();
  FadeOut();

  InitGameplayWorlds();
}

void GameLoop ()
{
  unsigned int NextCharIndex;
  unsigned int Lives;
  unsigned int ScoreAdderCounter;

  // Load main map/tileset
  LoadEfficientBGMap(kBGM_MainLR, kMNMain, 0);
  NextCharIndex = LoadEfficientCharset(kChr_Main, 0);
  SetBGMapPalette(kMNMain, BGM_PAL1);
  // Load tree map/tileset
  LoadEfficientBGMap(kBGM_DayTreesLR, kMNDayTrees, NextCharIndex);
  NextCharIndex = LoadEfficientCharset(kChr_DayTrees, NextCharIndex);
  // Load forest floor map/tileset
  LoadEfficientBGMap(kBGM_FloorLR, kMNFloor, NextCharIndex);
  NextCharIndex = LoadEfficientCharset(kChr_Floor, NextCharIndex);

  Level = 1 << 1;
  LevelStartScreen();

  EmptyBGMap(kMNStatus);

  Lives = 3;
  Score = 0;
  ScoreAdderCounter = 0;

  InitLevel();

  for(;;)
  {
    if(!NumBikers)
    {
      FadeOut();
      if(++Level >= (9 << 1))
        // Wrap around to level 1. Levels past 8 would be all the same
        // anyway since there aren't enough worlds to use for trees.
        Level = 1 << 1;
      Brightness = 32;
      StopAllSounds();
      AutoPause();
      LevelStartScreen();
      InitLevel();
    }

    PlayerControl();
    if(PlayerCrashed())
    {
      VIP_REGS[BKCOL] = 3;
      PlayNoiseGradient(75, 1500, 50, 0xFF, 0);
      while(NoiseFramesLeft)
      {
        WaitForVIP();
        DoSounds();
      }
      FadeOut();
      VIP_REGS[BKCOL] = 0;
      if(--Lives)
        InitLevel();
      else
        break;
    }
    ProcessThings();

    WaitForVIP();

    DoSounds();

    RenderThings();

    FloorMYOffset -= FloorMYOffsetDelta;
    if(FloorMYOffset < 0)
      FloorMYOffset = 7 << 3;
    RenderFloor();

    if(++ScoreAdderCounter == 25)
    {
      Score += PlayerSpeed;
      ScoreAdderCounter = 0;
    }

    PrintU32Big(kMNStatus, -32, 0, Lives);
    PrintU32Big(kMNStatus, 32, 0, Score);

    if(AutoPauseFrames)
      AutoPauseFrames--;
  }

  Brightness = 32;  // Can't stay dark if it was night
  StopAllSounds();
  SND_REGS[WAVE4].SxINT = 0;
}

int main ()
{
  LoadEfficientCharset(kChr_Font, 2048 - 64);  // The font has 64 characters

  vbDisplayOn();

  VIP_REGS[GPLT1] = VIP_REGS[JPLT1] = 0xE0;  // Palettes without color 1
  VIP_REGS[GPLT2] = VIP_REGS[JPLT2] = 0xD0;  // Palettes without color 2
  VIP_REGS[GPLT3] = VIP_REGS[JPLT3] = 0x90;  // Palettes without color 3

  Brightness = 32;

  #if kDebug
    AutoPauseOn = true;
    InitSound();
  #else
    PrecautionScreen();
    AdjustmentScreen();
    InitSound();
    AutoPauseSelect();
  #endif

  AutoPauseFrames = kAutoPausePeriod;
  HighScore = 0;

  for(;;)
  {
    TitleScreen();
    GameLoop();
    GameOver();
  }
}
