void PlaySound (const u16* DataPtr)
{
  // Sets up a sound to play. If there are too many sounds playing, the sound
  // will not play.

  register unsigned int i;

  for(i = 0; i < kMaxSounds; i++)
    if(!Sound[i].StartPtr)
    {
      Sound[i].StartPtr = (u16*)(DataPtr + 1);
      Sound[i].NoteIndex = -1;
      Sound[i].FramesPerNote = Sound[i].FramesLeft = *DataPtr;

      break;
    }
}

void StopSound (unsigned int SoundIndex)
{
  // Stops a sound.

  // Mark it as nonexistant.
  Sound[SoundIndex].StartPtr = NULL;
  
  // Make it silent.
  SND_REGS[SoundIndex].SxLRV = 0;
}

void StopAllSounds ()
{
  // Stops all sounds.

  register unsigned int i;

  for(i = 0; i < kMaxSounds; i++)
    StopSound(i);

  NoiseFramesLeft = 0;
}

void PauseAllSounds ()
{
  // Pauses all sounds by making them silent.

  SSTOP = 1;
}

void DoSounds ()
{
  // Changes notes on sounds if necessary or ends them.

  register unsigned int i;
  register unsigned short Note;

  for(i = 0; i < kMaxSounds; i++)
  {
    if(Sound[i].StartPtr)
    {
      if(!(--Sound[i].FramesLeft))
      {
        Sound[i].NoteIndex++;                          // Move to next note.
        Note = Sound[i].StartPtr[Sound[i].NoteIndex];  // Get it.

        // Set length of new note.
        Sound[i].FramesLeft = Sound[i].FramesPerNote;

        // Is it a special note?
        switch(Note)
        {
          case PAU:
            // Pause.
            SND_REGS[i].SxLRV = 0;
            continue;
        
          case HOLD:
            // Continue playing the previous note.
            continue;

          case ENDSOUND:
            // End the sound.
            StopSound(i);
            continue;

          case LOOPSOUND:
            // Restart the sound.
            Sound[i].NoteIndex = -1;  // Will get incremented next time.
            Sound[i].FramesLeft = 1;  // Will get decremented next time.
            continue;
        }

        // Send the note to the sound channel.
        SND_REGS[i].SxINT = 0x9F;         // I forgot what this is for.
        SND_REGS[i].SxFQH = Note >> 8;    // Frequency; higher 3 bits.
        SND_REGS[i].SxFQL = Note & 0xFF;  // Frequency; lower 8 bits.
        SND_REGS[i].SxLRV = 0xFF;         // Left/right speaker volume.
        SND_REGS[i].SxEV0 = 0xFC;         // No fadeout; volume is constant.
        SND_REGS[i].SxEV1 = 0x02;         // Repeat it forever.
      }
    }
  }

  // Noise gradient (TODO volume changing)
  if(NoiseFramesLeft)
  {
    NoiseFreq += NoiseFreqIncr;
    SND_REGS[NOISE].SxFQH = NoiseFreq >> 8;
    SND_REGS[NOISE].SxFQL = NoiseFreq & 0xFF;
    if(!(--NoiseFramesLeft))
      SND_REGS[NOISE].SxLRV = 0;
  }
}

void InitSound ()
{
  // Sets up the VSU for use with the sound engine.

  register unsigned int i;

  StopAllSounds();

  // Modulation data.
  for(i = 0; i <= 0x7C; i++) MODDATA[i << 2] = kModData[i];

  // Make a square waveform.
  for(i = 0; i <= 0x3C; i += 4) WAVEDATA1[i] = 0;        // The low part.
  for(i = 0x40; i <= 0x7C; i += 4) WAVEDATA1[i] = 0x3F;  // The high part.

  // Choose the same waveform for all channels.
  SND_REGS[WAVE1].SxRAM = 0;
  SND_REGS[WAVE2].SxRAM = 0;
  SND_REGS[WAVE3].SxRAM = 0;
  SND_REGS[WAVE4].SxRAM = 0;
}

void PlayNoiseGradient (
  u16 Duration, s16 StartFreq, s16 EndFreq, u8 StartVol, u8 EndVol
)
{
  register SOUNDREG* NoiseCh;

  NoiseFramesLeft = Duration;
  NoiseFreq = StartFreq;
  NoiseFreqIncr = (EndFreq - StartFreq) / NoiseFramesLeft;

  // TODO: maybe implement volume gradients in the future -MH

  NoiseCh = &(SND_REGS[NOISE]);
  NoiseCh->SxINT = 0x80;
  NoiseCh->SxFQH = StartFreq >> 8;
  NoiseCh->SxFQL = StartFreq & 0xFF;
  NoiseCh->SxLRV = StartVol;
  NoiseCh->SxEV0 = 0xF0;
  NoiseCh->SxEV1 = 0x30;
}
