unsigned int Abs (signed int X)
{
  // Returns the absolute value of an integer.

  return (X < 0) ? -X : X;
}

void CopyMem (void* Dst, void* Src, unsigned int Length)
{
  // Copies a block of bytes from Dst to Src.
  // The length must be nonzero.

  asm("_CopyMem_1:");
  asm("  ld.b 0[r7], r1");
  asm("  st.b r1, 0[r6]");
  asm("  add 1, r7");
  asm("  add 1, r6");
  asm("  add -1, r8");
  asm("  bne _CopyMem_1");
}

void Crash (char* Message)
{
  // Shows a message to the player and hangs.

  // Reload the font as it may have been overwritten
  LoadEfficientCharset(kChr_Font, 2048 - 64);  // The font has 64 characters

  EmptyBGMap(0);

  vbSetWorld(31, WRLD_ON | 0, 0, 0, 0, 0, 0, 0, 383, 223);
  SetEndWorld(30);

  vbTextOut(0, 0, 0, Message);

  vbDisplayShow();

  for(;;)
    asm("halt");
}

void NumberStrU32 (char* Buffer, u32 What)
{
  // Converts an unsigned 32-bit integer into decimal representation and writes
  // it to a buffer. The buffer must be at least 11 bytes long.

  // Based on the itoa function from the libgccvb that's distributed with
  // Fwirt's 6 demos.

  register unsigned int i;

  // The actual processing.
  for(i = 0; i < 10; i++)
  {
    Buffer[9 - i] = What % 10 + '0';
    What /= 10;
  }

  // Replace leading zeros with spaces for aesthetic quality.
  for(i = 0; i < 9; i++)
    if(Buffer[i] == '0')
      Buffer[i] = ' ';
    else
      break;

  // Null-terminate it.
  Buffer[10] = 0;
}

void PrintU32Big (
  unsigned int BGMap, unsigned int X, unsigned int Y, u32 What
)
{
  // Prints an unsigned 32-bit integer to a BGMap in decimal representation with
  // 16*16-pixel numbers.

  char Buffer[11];
  register unsigned int i;
  register unsigned int Digit;
  register u16* BGMap0 = (u16*)0x20000;
  register u16* DstPtr;

  NumberStrU32(Buffer, What);

  DstPtr = (u16*)(0x20000 + BGMap * 0x2000 + Y * 128 + X);

  i = 0;
  while(Buffer[i])
  {
    // Source BGMap cells are on rows 32 and 33.

    Digit = (Buffer[i] - '0') * 2;

    DstPtr[0] = BGMap0[32 * 64 + Digit];
    DstPtr[1] = BGMap0[32 * 64 + Digit + 1];
    DstPtr[64] = BGMap0[33 * 64 + Digit];
    DstPtr[65] = BGMap0[33 * 64 + Digit + 1];

    DstPtr += 2;
    i++;
  }
}

void PrintU32 (
  unsigned int BGMap, unsigned int X, unsigned int Y, u32 What
)
{
  // Prints an unsigned 32-bit integer to a BGMap in decimal representation
  // with vbTextOut.

  char Buffer[11];

  NumberStrU32(Buffer, What);
  vbTextOut(BGMap, X, Y, Buffer);
}

void PrintS32 (
  unsigned int BGMap, unsigned int X, unsigned int Y, s32 What
)
{
  // Prints a signed 32-bit integer to a BGMap in decimal representation with
  // vbTextOut.

  if(What < 0)
  {
    vbCharOut(BGMap, X, Y, '-');
    What = -What;
  }
  else
    vbCharOut(BGMap, X, Y, ' ');

  PrintU32(BGMap, X + 1, Y, What);
}

void SetMem (void* Dst, unsigned int Value, unsigned int Length)
{
  // Sets a block of bytes to a given value.
  // The length must be nonzero.

  asm("_SetMem_1:");
  asm("  st.b r7, 0[r6]");
  asm("  add 1, r6");
  asm("  add -1, r8");
  asm("  bne _SetMem_1");
}
