// libgccvb replacement library
// Not a drop-in replacement for libgccvb, but a minimalistic library with a
// subset of its features.

// VIP world structure
typedef struct
{
  u16 Flags;
  s16 GX;
  s16 GP;
  s16 GY;
  s16 MX;
  s16 MP;
  s16 MY;
  u16 Width;
  u16 Height;
  u16 ParAddr;
  u16 OOBChar;
  u16 Reserved[5];
} vbtyWorld;

// VIP affine parameter structure
typedef struct
{
  s16 MX;  // 13.3 fixed point
  s16 MP;
  s16 MY;  // 13.3 fixed point
  s16 XIncr;  // 7.9 fixed point
  s16 YIncr;  // 7.9 fixed point
  u16 Reserved[3];
} vbtyAffinePar;

// Pointer to VIP character memory
#define vbCharmem ((u16*)0x78000)

// Pointers to VIP character segments
#define vbCharseg0 ((u16*)0x78000)  // Characters 0 to 511
#define vbCharseg1 ((u16*)0x7A000)  // Characters 512 to 1023
#define vbCharseg2 ((u16*)0x7C000)  // Characters 1024 to 1535
#define vbCharseg3 ((u16*)0x7E000)  // Characters 1536 to 2047

// Macro to return the address of a character by index (0 to 2047)
#define vbCharAddress(c) (&(((u16*)0x78000)[c << 3]))

// Pointer to VIP world array
#define vbWorld ((vbtyWorld*)0x3D800)

#ifndef BGM_CSET
  // Aliases for libgccvb macros so the old names from gccVB 1.0 can be used.

  #define BGM_CSET(n, x, y, c) BGMM[((n) << 12) + ((y) << 6) + (x)] = (BGMM[((n) << 12) + ((y) << 6) + (x)] & 0xF000) | (u16)(c)

  #define BGM_PAL0 0x0000
  #define BGM_PAL1 0x4000
  #define BGM_PAL2 0x8000
  #define BGM_PAL3 0xC000

  #define OBJ_SET_X OBJ_XSET
  #define OBJ_SET_Y OBJ_YSET
  #define OBJ_SET_P OBJ_PSET
  #define OBJ_SET_PAL OBJ_PALSET
  #define OBJ_SET_CHAR OBJ_CSET
  #define OBJ_SET_VIS OBJ_VIS
#endif

// The following definitions and declarations are based on libgccvb:

#define BGMM ((u16*)0x20000)
#define WAM ((u16*)0x3D800)

// Hardware reg SCR definitions
#define	S_INTDIS 0x80  // Disable Interrups
#define	S_SW     0x20  // Software Reading
#define	S_SWCK   0x10  // Software Clock, Interrupt?
#define	S_HW     0x04  // Hardware Reading
#define	S_STAT   0x02  // Hardware Reading Status
#define	S_HWDIS  0x01  // Disable Hardware Reading

// Controller definitions
#define	K_PWR 0x0001  // Low Battery
#define	K_A   0x0004  // A Button
#define	K_B   0x0008  // B Button
#define	K_RT  0x0010  // R Trigger
#define	K_LT  0x0020  // L Trigger
#define	K_RU  0x0040  // Right Pad, Up
#define	K_RR  0x0080  // Right Pad, Right
#define	K_LR  0x0100  // Left Pad, Right
#define	K_LL  0x0200  // Left Pad, Left
#define	K_LD  0x0400  // Left Pad, Down
#define	K_LU  0x0800  // Left Pad, Up
#define	K_STA 0x1000  // Start Button
#define	K_SEL 0x2000  // Select Button
#define	K_RL  0x4000  // Right Pad, Left
#define	K_RD  0x8000  // Right Pad, Down

#define	WRLD_ON	    0xC000
#define	WRLD_LON    0x8000
#define	WRLD_RON    0x4000
#define	WRLD_OBJ    0x3000
#define	WRLD_AFFINE 0x2000
#define	WRLD_HBIAS  0x1000
#define	WRLD_OVR    0x0080
#define	WRLD_END    0x0040

#define	WORLD_HEAD(n, head) WAM[(n) << 4] = head

// The following functions are based on libgccvb:

void vbDisplayOn ()
{
  VIP_REGS[REST] = 0;
  VIP_REGS[XPCTRL] = VIP_REGS[XPSTTS] | XPEN;
  VIP_REGS[DPCTRL] = VIP_REGS[DPSTTS] | SYNCE | RE | DISP;
  VIP_REGS[FRMCYC] = 0;
  VIP_REGS[INTCLR] = VIP_REGS[INTPND];

  VIP_REGS[BRTA] = 0;
  VIP_REGS[BRTB] = 0;
  VIP_REGS[BRTC] = 0;
  VIP_REGS[GPLT0] = 0xE4;
  VIP_REGS[GPLT1] = 0xE4;
  VIP_REGS[GPLT2] = 0xE4;
  VIP_REGS[GPLT3] = 0xE4;
  VIP_REGS[JPLT0] = 0xE4;
  VIP_REGS[JPLT1] = 0xE4;
  VIP_REGS[JPLT2] = 0xE4;
  VIP_REGS[JPLT3] = 0xE4;
  VIP_REGS[BKCOL] = 0;
}

void vbDisplayShow ()
{
  VIP_REGS[BRTA] = 32;
  VIP_REGS[BRTB] = 64;
  VIP_REGS[BRTC] = 32;
}

u16 vbReadPad ()
{
  HW_REGS[SCR] = S_INTDIS | S_HW;
  while(HW_REGS[SCR] & S_STAT);
  return (HW_REGS[SDHR] << 8) | HW_REGS[SDLR];
}

void vbSetWorld (
  unsigned int Which, u16 Flags, s16 GX, s16 GP, s16 GY, u16 MX, s16 MP,
  u16 MY, u16 Width, u16 Height
)
{
  register vbtyWorld* World;

  World = &(vbWorld[Which]);

  World->Flags = Flags;
  World->GX = GX;
  World->GP = GP;
  World->GY = GY;
  World->MX = MX;
  World->MP = MP;
  World->MY = MY;
  World->Width = Width;
  World->Height = Height;
}

void vbTextOut (u16 BGMap, u16 X, u16 Y, char* What)
{
  register u16 i;
  register unsigned int Pos;

  i = 0;
  Pos = 0x1000 * BGMap + Y * 64 + X;

  while(What[i])
  {
    BGMM[Pos + i] = 2048 - 64 + What[i] - 32;  // 64 chars, starting at 32
    i++;
  }
}